#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>
#include <unistd.h>
#include <errno.h>
#ifdef _OPENMP
#include <pthread.h>
#endif
#include "libtriplex/gff3.h"
#include "common/helpers.h"

#define GFF3_SCACHE_VALMAXLEN 2

static struct gff3_blk *gff3_blk_new(size_t blk_size)
{
	struct gff3_blk *b;

	b = malloc(sizeof(struct gff3_blk) + (sizeof(gff3_rec_t) * blk_size));

	if (b == NULL)
		return (NULL);

	b->next     = NULL;
	b->blk_size = blk_size;
	b->blk_next = 0;

	memset(b->blk_item, 0, sizeof(gff3_rec_t) * blk_size);

	return (b);
}

static void gff3_rec_free_content(gff3_rec_t *r)
{
	register uint16_t i;

        for (i = 0; i < r->attr_cnt; ++i) {
                if (strlen(r->attr_tbl[i].value) > GFF3_SCACHE_VALMAXLEN)
                        free(r->attr_tbl[i].value);
        }

	free(r->attr_tbl);

	return;
}

static void gff3_blk_free(struct gff3_blk *blk)
{
	size_t bsz, i;

	bsz = sizeof(struct gff3_blk) + (sizeof(gff3_rec_t) * blk->blk_size);

	for (i = 0; i < blk->blk_next; ++i)
		gff3_rec_free_content(blk->blk_item + i);

	memset(blk, 0, bsz);
	free(blk);

	return;
}

gff3_t *gff3_new(const char *source)
{
	gff3_t *g = alloc_type(gff3_t);

	if (g == NULL)
		return (NULL);

        g->source = strdup(source);
	g->blk0   = gff3_blk_new(sysconf(_SC_PAGESIZE)/sizeof(gff3_rec_t));
	g->blkN   = g->blk0;

        g->scache = NULL;
        g->scount = 0;

	if (g->blk0 == NULL) {
		free(g);
		return (NULL);
	}

	g->rcnt = 0;
#ifdef _OPENMP
        pthread_rwlock_init(&g->rwlock, NULL);
#endif
	return (g);
}

gff3_rec_t *gff3_rec_prepare(gff3_rec_t *r, uint16_t attr_cnt)
{
	r->seq_id = NULL;
	r->type   = NULL;
	r->start  = 0;
	r->end    = 0;
	r->score  = 0.0;
	r->strand = '+';
	r->phase  =  -1;

	r->attr_cnt = 0;
	r->attr_tbl = NULL;

	return (r);
}

int gff3_rec_addtag(gff3_rec_t *rec, const char *tag, char *value)
{
	rec->attr_tbl = realloc_array(rec->attr_tbl, gff3_attr_t, ++rec->attr_cnt);

	rec->attr_tbl[rec->attr_cnt - 1].tag   = (char *)tag;
	rec->attr_tbl[rec->attr_cnt - 1].value = value;

	return (0);
}

static int scache_bscmp(const char *key, const char **str)
{
        return strcmp(key, *str);
}

static int scache_qscmp(const char **a, const char **b)
{
        return strcmp(*a, *b);
}

static char *scache_get(gff3_t *g, const char *str)
{
        char *res, **ref = bsearch(str, g->scache, g->scount, sizeof(char *),
                             (int(*)(const void *, const void *))scache_bscmp);

        if (ref == NULL) {
                g->scache = realloc_array(g->scache, char *, ++g->scount);
                g->scache[g->scount - 1] = res = strdup(str);

                qsort(g->scache, g->scount, sizeof(char *),
                      (int(*)(const void *, const void *))scache_qscmp);
        } else
                res = *ref;

        return (res);
}

int gff3_rec_add(gff3_t *g, gff3_rec_t *rec)
{
	struct gff3_blk *dst;

	if (g == NULL) {
		errno = EFAULT;
		return -1;
	}

#ifdef _OPENMP
        pthread_rwlock_wrlock(&g->rwlock);
#endif
	if (g->blk0 == NULL) {
		g->blk0 = gff3_blk_new(GFF3_DEFAULT_BLKSIZE);
		g->blkN = g->blk0;
		dst     = g->blk0;

		goto blk_add;
	}

	dst = g->blk0;

	for (;;) {
		if (dst->blk_next < dst->blk_size)
			goto blk_add;

		if (dst->next == NULL) {
			dst->next = gff3_blk_new(GFF3_DEFAULT_BLKSIZE);
			g->blkN   = dst->next;
			dst       = g->blkN;

			goto blk_add;
		}

		dst = dst->next;
	}
#ifdef _OPENMP
        pthread_rwlock_unlock(&g->rwlock);
#endif
	return (-1);
blk_add:
	if (dst == NULL) {
                abort ();
        } else {
                register uint16_t i;
                gff3_rec_t *recdst = dst->blk_item + dst->blk_next;

                memcpy (recdst, rec, sizeof(gff3_rec_t));

                ++dst->blk_next;
                ++g->rcnt;

                recdst->seq_id = scache_get(g, rec->seq_id);
                recdst->type   = scache_get(g, rec->type);

                for (i = 0; i < recdst->attr_cnt; ++i) {
                        recdst->attr_tbl[i].tag   = scache_get(g, recdst->attr_tbl[i].tag);

                        /*
                         * For tag values, cache only string of length GFF3_SCACHE_VALMAXLEN or less.
                         */
                        if (strlen(recdst->attr_tbl[i].value) <= GFF3_SCACHE_VALMAXLEN) {
                                char *tmp;
                                recdst->attr_tbl[i].value = scache_get(g, tmp = recdst->attr_tbl[i].value);
                                free(tmp);
                        }
                }
        }

#ifdef _OPENMP
        pthread_rwlock_unlock(&g->rwlock);
#endif
	return (0);
}

static int gff3_write_hdr(FILE *fp)
{
	rewind(fp);
	return !ferror(fp) ?
		fputs("##gff-version 3\n", fp) : -1;
}

static int gff3_write_rec(FILE *fp, char *source, gff3_rec_t *rec)
{
	char phase;

	switch (rec->phase) {
	case 0:  phase = '0'; break;
	case 1:  phase = '1'; break;
	case 2:	 phase = '2'; break;
	default: phase = '.';
	}

	fprintf(fp,
		"%s\t"         /* sequence ID */
		"%s\t"         /* source */
		"%s\t"         /* type (SOFA term or SOFA access #) */
		"%"PRIu64"\t"  /* start */
		"%"PRIu64"\t"  /* end */
		"%g\t"         /* score */
		"%c\t"         /* strand */
		"%c\t",        /* phase */
		rec->seq_id,
		source,
		rec->type,
		rec->start,
		rec->end,
		rec->score,
		rec->strand,
		phase);

	if (rec->attr_cnt > 0) {
		register uint16_t i;

		for (i = 0; i < (rec->attr_cnt - 1); ++i)
			fprintf (fp, "%s=%s;", rec->attr_tbl[i].tag, rec->attr_tbl[i].value);
		if ((rec->attr_cnt - i) > 0)
			fprintf (fp, "%s=%s",  rec->attr_tbl[i].tag, rec->attr_tbl[i].value);
	} else
		fputc('.', fp);

	fprintf(fp, "\n");

	return (0);
}

int gff3_write(gff3_t *g, const char *path)
{
	register struct gff3_blk *blk, *old;
	register size_t i;
        FILE *fp;
        int ret = 0;

	if (g == NULL) {
		errno = EFAULT;
		return (-1);
	}

        if (path == NULL)
		fp = stdout;
	else if (strcmp(path, "-") == 0)
                fp = stdout;
        else
		fp = fopen(path, "w");

	if (fp == NULL)
		return (-1);

        gff3_write_hdr(fp);

#ifdef _OPENMP
        pthread_rwlock_rdlock(&g->rwlock);
#endif
	blk = g->blk0;

	while (blk != NULL) {
		for (i = 0; i < blk->blk_next; ++i) {
			if (gff3_write_rec(fp, g->source, blk->blk_item + i) < 0) {
                                ret = -1;
                                goto _ret;
                        }
			/* TODO: add blk_offset for better error recovery */
		}

		old = blk;
		blk = blk->next;

		fflush(fp);
	}
_ret:
#ifdef _OPENMP
        pthread_rwlock_unlock(&g->rwlock);
#endif
        if (path != NULL)
                fclose(fp);

	return(ret);
}

int gff3_free(gff3_t *g)
{
        struct gff3_blk *blk, *blk_next;
        size_t i;

        blk = g->blk0;

        while (blk != NULL) {
                blk_next = blk->next;
                gff3_blk_free(blk);
                blk = blk_next;
        }

        if (g->scache != NULL) {
                for (i = 0; i < g->scount; ++i)
                        free(g->scache[i]);
                free(g->scache);
        }

        free(g->source);

#ifdef _OPENMP
        pthread_rwlock_destroy(&g->rwlock);
#endif
        free(g);

	return (0);
}
