/*
 *  Copyright (C) 2011 Masaryk University, Faculty of Informatics
 *
 *  Author: Daniel Kopecek <xkopecek@fi.muni.cz>
 *
 *  LICENSE TERMS
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *  3. Neither the name of the Company nor the names of its contributors
 *  may be used to endorse or promote products derived from this
 *  software without specific prior written permission.
 *
 *  This software is provided ``as is'', and any express or implied
 *  warranties, including, but not limited to, the implied warranties of
 *  merchantability and fitness for a particular purpose are disclaimed.
 *  In no event shall the company or contributors be liable for any
 *  direct, indirect, incidental, special, exemplary, or consequential
 *  damages (including, but not limited to, procurement of substitute
 *  goods or services; loss of use, data, or profits; or business
 *  interruption) however caused and on any theory of liability, whether
 *  in contract, strict liability, or tort (including negligence or
 *  otherwise) arising in any way out of the use of this software, even
 *  if advised of the possibility of such damage.
 */
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>
#include <math.h>
#include <errno.h>
#include <libconfig.h>
#include <libfasta/fasta.h>

#include "libtriplex/gff3.h"
#include "common/helpers.h"
#include "libtriplex_private.h" /* includes the public header */
#include "algorithm.h"

const int  default_tri_types[] = { 0 };
const size_t default_tri_count = 1;

const triplexParam default_params[] = {
        { "triplex-type",    "type of the triplexes to be searched for", TRIPLEX_PARAM_INT|TRIPLEX_PFLAG_ARRAY, .value._array= { .count = 1, .values._int= (int *)default_tri_types } },
        { "max-steps",       "maximum steps of the DP algorithm",        TRIPLEX_PARAM_INT, .value._int= 200 },
        { "min-length",      "minimum length of a triplex",              TRIPLEX_PARAM_INT, .value._int= 6 },
        { "max-length",      "maximum length of a triplex",              TRIPLEX_PARAM_INT, .value._int= 200 },
        { "min-score",       "minimum DP score of a triplex",            TRIPLEX_PARAM_INT, .value._int= 3 },
        { "min-loop-length", "minimum loop length in a triplex",         TRIPLEX_PARAM_INT, .value._int= 3 },
        { "max-loop-length", "maximum loop length in a triplex",         TRIPLEX_PARAM_INT, .value._int= 10 },
        { "lambda_0145",     "lambda for triplex types 0,1,4,5",         TRIPLEX_PARAM_FLOAT, .value._float= 0.68 },
        { "lambda_2367",     "lambda for triplex types 2,3,6,7",         TRIPLEX_PARAM_FLOAT, .value._float= 0.68 },
        { "mi_0145",         "mi for triplex types 0,1,4,5",             TRIPLEX_PARAM_FLOAT, .value._float= 5.98 },
        { "mi_2367",         "mi for triplex types 2,3,6,7",             TRIPLEX_PARAM_FLOAT, .value._float= 5.97 },
        { "pen-isochange",   "penalization for leaving an isogroup",     TRIPLEX_PARAM_INT, .value._int= 5 },
        { "pen-mismatch",    "penalization for a mismatch",              TRIPLEX_PARAM_INT, .value._int= 5 },
        { "pen-insdel",      "penalization for insertion/deletion",      TRIPLEX_PARAM_INT, .value._int= 5 },
        { "score-isostay",   "score increment for staying in the same isogroup", TRIPLEX_PARAM_INT, .value._int = 0 },
        { "dtwist-limit",    "",                                         TRIPLEX_PARAM_INT, .value._int= 15 },
        { "p-value",         "report only triplexes with p_value(t) <= p_val", TRIPLEX_PARAM_FLOAT, .value._float= 2.0 },
        { "max-segment",     "maximum length of a sequence segment analyzed at a time", TRIPLEX_PARAM_INT, .value._int= 6*1024*1024 }
};

#define default_params_count (sizeof default_params / sizeof(triplexParam))

static int triplexParamCmp(const triplexParam *a, const triplexParam *b)
{
        return strcmp(a->name, b->name);
}

static int triplexParamKeyCmp(const char *key, const triplexParam *p)
{
        return strcmp(key, p->name);
}

#ifndef NDEBUG
static void print_params(const triplexParam *p, size_t c)
{
        size_t i;
        char *type;

        fprintf(stderr, "===== params =====\n");

        for (i = 0; i < c; ++i) {
                fprintf(stderr, "  Name: %s\n", p[i].name);
                fprintf(stderr, "  Desc: %s\n", p[i].description);
                switch(p[i].flags & TRIPLEX_PARAM_TYPEMASK) {
                case TRIPLEX_PARAM_INT:
                        type = "integer";
                        break;
                case TRIPLEX_PARAM_FLOAT:
                        type = "float";
                        break;
                case TRIPLEX_PARAM_STRING:
                        type = "string";
                        break;
                default:
                        type = "unknown";
                }

                fprintf(stderr, " Flags: %08x (Type: %s)\n", p[i].flags, type);
                fprintf(stderr, "\n");
        }

        fprintf(stderr, "==================\n");
}
#endif

triplexCTX *libtriplex_init(void)
{
        triplexCTX *ctx;

        ctx = alloc_type(triplexCTX);

        if (ctx == NULL)
                return (NULL);

        ctx->params = alloc_array(triplexParam, default_params_count);
        ctx->pcount = default_params_count;

        if (ctx->params == NULL) {
                errno_protect {
                        free(ctx);
                }
                return (NULL);
        }

        memcpy(ctx->params, default_params, sizeof(triplexParam) * default_params_count);

        /*
         * Sort the parameters so that we can use bsearch() to search for them later
         */
        qsort(ctx->params, default_params_count, sizeof(triplexParam),
              (int(*)(const void *, const void *))triplexParamCmp);

#ifndef NDEBUG
        print_params(ctx->params, ctx->pcount);
#endif
        ctx->flags = 0;

        return (ctx);
}

int libtriplex_setflags(triplexCTX *ctx, uint32_t flags)
{
        ctx->flags |= flags;
        return (0);
}

int libtriplex_unsetflags(triplexCTX *ctx, uint32_t flags)
{
        ctx->flags &= ~flags;
        return (0);
}

void libtriplex_fini(triplexCTX *ctx)
{
        /* Free params if needed */
        for (; ctx->pcount > 0; --ctx->pcount) {
                triplexParam *p = ctx->params + ctx->pcount - 1;

                if (p->flags & TRIPLEX_PFLAG_FREE) {
                        if (p->flags & TRIPLEX_PFLAG_ARRAY) {
                                switch(p->flags & TRIPLEX_PARAM_TYPEMASK) {
                                case TRIPLEX_PARAM_INT:
                                        free(p->value._array.values._int);
                                        break;
                                case TRIPLEX_PARAM_FLOAT:
                                        free(p->value._array.values._float);
                                        break;
                                case TRIPLEX_PARAM_STRING:
                                        while(p->value._array.count > 0) {
                                                free(p->value._array.values._string[p->value._array.count - 1]);
                                                --p->value._array.count;
                                        }

                                        free(p->value._array.values._string);
                                        break;
                                default:
                                        abort();
                                }
                        } else
                                switch(p->flags & TRIPLEX_PARAM_TYPEMASK) {
                                case TRIPLEX_PARAM_INT:
                                case TRIPLEX_PARAM_FLOAT:
                                        break;
                                case TRIPLEX_PARAM_STRING:
                                        free(p->value._string);
                                        break;
                                default:
                                        abort();
                                }
                }
        }

        if (ctx->params != NULL)
                free(ctx->params);

        /* Finally free the structure */
        free(ctx);
}

int libtriplex_setparam(triplexCTX *ctx, const char *name, uint32_t type, ...)
{
        triplexParam *p;
        va_list ap;

        p = bsearch(name, ctx->params, ctx->pcount, sizeof(triplexParam),
                    (int(*)(const void *, const void *))triplexParamKeyCmp);

        if (p == NULL)
                return (LIBTRIPLEX_EINVALP);

        if ((p->flags & (TRIPLEX_PARAM_TYPEMASK|TRIPLEX_PFLAG_ARRAY)) != type)
                return (LIBTRIPLEX_EINVALV);

        va_start(ap, type);

        if (type & TRIPLEX_PFLAG_ARRAY) {
                switch(type & TRIPLEX_PARAM_TYPEMASK) {
                case TRIPLEX_PARAM_INT:
                        p->value._array.values._int = va_arg(ap, int *);
                        p->value._array.count = va_arg(ap, size_t);
                        break;
                case TRIPLEX_PARAM_FLOAT:
                        p->value._array.values._float = va_arg(ap, double *);
                        p->value._array.count = va_arg(ap, size_t);
                        break;
                case TRIPLEX_PARAM_STRING:
                        p->value._array.values._string = va_arg(ap, char **);
                        p->value._array.count = va_arg(ap, size_t);
                        break;
                }
        } else {
                switch(type & TRIPLEX_PARAM_TYPEMASK) {
                case TRIPLEX_PARAM_INT:
                        p->value._int = va_arg(ap, int);
                        break;
                case TRIPLEX_PARAM_FLOAT:
                        p->value._float = va_arg(ap, double);
                        break;
                case TRIPLEX_PARAM_STRING:
                        p->value._string = va_arg(ap, char *);
                        break;
                }
        }

        va_end(ap);
        return (0);
}

int libtriplex_setparam_coerce(triplexCTX *ctx, const char *name, const char *value)
{
        triplexParam *p;
        va_list ap;

        dP("DEBUG: setparam_coerce: %s = %s\n", name, value);

        p = bsearch(name, ctx->params, ctx->pcount, sizeof(triplexParam),
                    (int(*)(const void *, const void *))triplexParamKeyCmp);

        if (p == NULL)
                return (LIBTRIPLEX_EINVALP);

        if (p->flags & TRIPLEX_PFLAG_ARRAY) {
                switch(p->flags & TRIPLEX_PARAM_TYPEMASK) {
                case TRIPLEX_PARAM_INT: {
                        int   *array = NULL;
                        size_t count = 0;
                        char  *savep = NULL, *tok, *str;

                        dP("DEBUG:  array of integers\n");

                        str = strdup(value);
                        tok = strtok_r(str, ",", &savep);

                        do {
                                array = realloc_array(array, int, ++count);
                                array[count - 1] = strtol(tok, NULL, 10);
                                dP("DEBUG:    a[%zu] = %d\n", count - 1, array[count - 1]);
                        } while ((tok = strtok_r(NULL, ",", &savep)) != NULL);

                        p->value._array.count       = count;
                        p->value._array.values._int = array;

                        free(str);
                }       break;
                case TRIPLEX_PARAM_FLOAT: {
                        double *array = NULL;
                        size_t  count = 0;
                        char   *savep = NULL, *tok, *str;

                        dP("DEBUG:  array of floating point numbers\n");

                        str = strdup(value);
                        tok = strtok_r(str, ",", &savep);

                        do {
                                array = realloc_array(array, double, ++count);
                                array[count - 1] = strtod(tok, NULL);
                                dP("DEBUG:    a[%zu] = %f\n", count - 1, array[count - 1]);
                        } while ((tok = strtok_r(NULL, ",", &savep)) != NULL);

                        p->value._array.count         = count;
                        p->value._array.values._float = array;

                        free(str);
                }       break;
                case TRIPLEX_PARAM_STRING: {
                        char  **array = NULL;
                        size_t  count = 0;
                        char   *savep = NULL, *tok, *str;

                        dP("DEBUG:  array of strings\n");

                        str = strdup(value);
                        tok = strtok_r(str, ",", &savep);

                        do {
                                array = realloc_array(array, char *, ++count);
                                array[count - 1] = strdup(tok);
                                dP("DEBUG:    a[%zu] = %s\n", count - 1, array[count - 1]);
                        } while ((tok = strtok_r(NULL, ",", &savep)) != NULL);

                        p->value._array.count          = count;
                        p->value._array.values._string = array;

                        free(str);
                }       break;
                }
        } else {
                switch(p->flags & TRIPLEX_PARAM_TYPEMASK) {
                case TRIPLEX_PARAM_INT:
                        p->value._int = strtol(value, NULL, 10);
                        dP("DEBUG:  integer = %d\n", p->value._int);
                        break;
                case TRIPLEX_PARAM_FLOAT:
                        p->value._float = strtod(value, NULL);
                        dP("DEBUG:  floating point number = %f\n", p->value._float);
                        break;
                case TRIPLEX_PARAM_STRING:
                        p->value._string = strdup(value);
                        dP("DEBUG:  string = %s\n", p->value._string);
                        break;
                }
        }

        va_end(ap);
        return (0);
}

int libtriplex_getparam(triplexCTX *ctx, const char *name, uint32_t type, ...)
{
        triplexParam *p;
        va_list ap;

        p = bsearch(name, ctx->params, ctx->pcount, sizeof(triplexParam),
                    (int(*)(const void *, const void *))triplexParamKeyCmp);

        if (p == NULL)
                return (LIBTRIPLEX_EINVALP);

        /* if the given and defined types don't match, return EINVAL */
        if ((p->flags & (TRIPLEX_PARAM_TYPEMASK|TRIPLEX_PFLAG_ARRAY)) != type)
                return (LIBTRIPLEX_EINVALV);

        va_start(ap, type);

        if (type & TRIPLEX_PFLAG_ARRAY) {
                switch(type & TRIPLEX_PARAM_TYPEMASK) {
                case TRIPLEX_PARAM_INT: {
                        int   **dst_a = va_arg(ap, int **);
                        size_t *dst_c = va_arg(ap, size_t *);

                        *dst_a = p->value._array.values._int;
                        *dst_c = p->value._array.count;
                }       break;
                case TRIPLEX_PARAM_FLOAT: {
                        double **dst_a = va_arg(ap, double **);
                        size_t  *dst_c = va_arg(ap, size_t *);

                        *dst_a = p->value._array.values._float;
                        *dst_c = p->value._array.count;
                }       break;
                case TRIPLEX_PARAM_STRING: {
                        char  ***dst_a = va_arg(ap, char ***);
                        size_t  *dst_c = va_arg(ap, size_t *);

                        *dst_a = p->value._array.values._string;
                        *dst_c = p->value._array.count;
                } break;
                }
        } else {
                switch(type & TRIPLEX_PARAM_TYPEMASK) {
                case TRIPLEX_PARAM_INT: {
                        int *dst = va_arg(ap, int *);
                        *dst = p->value._int;
                }       break;
                case TRIPLEX_PARAM_FLOAT: {
                        double *dst = va_arg(ap, double *);
                        *dst = p->value._float;
                }       break;
                case TRIPLEX_PARAM_STRING: {
                        char **dst = va_arg(ap, char **);
                        *dst = p->value._string;
                }       break;
                }
        }

        va_end(ap);

        return (0);
}

int libtriplex_search(triplexCTX *ctx, const char *path, gff3_t *result)
{
        int tri_type, ret = 0;
        FASTA *fa;
        atrans_t *tr;

        FASTA_rec_t farec_mem, *farec;

        uint32_t faread_flags, faopen_flags;
        uint32_t seq_maxlen, seg_maxlen;

        int   *tri_types, all_tri_types[8] = {0, 1, 2, 3, 4, 5, 6, 7 };
        size_t tri_count = 0, t = 0;

        if (ctx == NULL || path == NULL || result == NULL)
                return (LIBTRIPLEX_EFAULT);

        /* Open flags:
         *
         * FASTA_READ     - open for reading
         * FASTA_ONDEMSEQ - load sequence data on demand (see read flags bellow)
         * FASTA_NASEQ    - expect the sequence data to be a nucleic acid sequence
         * FASTA_USEINDEX - use FASTA index file if available
         * FASTA_GENINDEX - generate the index if not available
         * FASTA_CHKINDEX - perform minimal checks on the index to ensure that it is up-to-date.
         */
        faopen_flags = FASTA_READ|FASTA_ONDEMSEQ|FASTA_NASEQ;

        if (!(ctx->flags & LIBTRIPLEX_FLAG_IDXOFF))
                faopen_flags = FASTA_USEINDEX|FASTA_GENINDEX|FASTA_CHKINDEX;

        /* Read flags:
         *
         * FASTA_INMEMSEQ - load the sequence data into memory
         * FASTA_MAPCDSEG - map coding sequences
         */
        faread_flags = FASTA_INMEMSEQ;

        if (!(ctx->flags & LIBTRIPLEX_FLAG_CDSOFF))
                faread_flags |= FASTA_MAPCDSEG;

        if (result == NULL)
                return (LIBTRIPLEX_EFAULT);

        libtriplex_getparam(ctx, "max-length",   TRIPLEX_PARAM_INT, &seq_maxlen);
        libtriplex_getparam(ctx, "max-segment",  TRIPLEX_PARAM_INT, &seg_maxlen);
        libtriplex_getparam(ctx, "triplex-type",
                            TRIPLEX_PARAM_INT|TRIPLEX_PFLAG_ARRAY, &tri_types, &tri_count);

        if (tri_count == 1 && tri_types[0] == 8) {
                tri_types = all_tri_types;
                tri_count = sizeof all_tri_types/sizeof(int);
        }

	/*
	 * Setup alphabet translation
	 */
	tr = atrans_new(8, 8, 0, 0);
	tr->tr_letter_s2d['A'] = tr->tr_letter_s2d['a'] = 0;
	tr->tr_letter_s2d['C'] = tr->tr_letter_s2d['c'] = 1;
	tr->tr_letter_s2d['G'] = tr->tr_letter_s2d['g'] = 2;
	tr->tr_letter_s2d['T'] = tr->tr_letter_s2d['t'] = 3;

        /*
         * Open the FASTA (sequence data) source
         */
        fa = fasta_open(path, faopen_flags, tr);

        if (fa == NULL)
                return (LIBTRIPLEX_ESOURCE);

        /*
         * Load a record from the FASTA source into memory
         */
        while ((farec = fasta_read(fa, &farec_mem, faread_flags, NULL)) != NULL && ret == 0) {
                FASTA_CDS_t facds_mem, *facds;
                int search_ret;

                dP("DEBUG: farec=%p, id=%s\n", farec, farec->rec_id);
                /*
                 * Search for triplexes in each segment
                 */
                while ((facds = fasta_read_CDS(fa, farec, &facds_mem, 0)) != NULL && ret == 0) {
                        register uint64_t s, k, k_max;

                        dP("DEBUG:  facds=%p, seg_len=%"PRIu64"\n", facds, facds->seg_len);

                        /*
                         * Compute number of steps and handle overflows
                         * (max-length <= max-segment <= sequence-length)
                         */
                        k_max = 1 + llround((double)(facds->seg_len -
                                                     (seg_maxlen <= facds->seg_len ?
                                                      seg_maxlen : facds->seg_len))/(double)(seg_maxlen - seq_maxlen + 1));
                        /*
                         * For each segment, search in a sliding window using
                         * the max_length parameter as the window width
                         */
#pragma omp parallel for private(search_ret, s, k) shared(ret)
                        for (k = 0; k < k_max; ++k) {
                                s = k * (seg_maxlen - seq_maxlen + 1);
                                dP("DEBUG:   s=%"PRIu64"\n", s);
#pragma omp flush(ret)
                                for(t = 0; t < tri_count && ret == 0; ++t) {
                                        search_ret = triplex_search(ctx, farec->rec_id, facds->seg_mem + s,
                                                                    /*
                                                                     * If there is > seg_maxlen left, use seg_maxlen.
                                                                     * Otherwise compute the length of the last segment
                                                                     * of the sequence to be analyzed.
                                                                     */
                                                                    s + seg_maxlen < facds->seg_len ? seg_maxlen : facds->seg_len - s,
                                                                    /*
                                                                     * Set offset only if CDS mapping is on.
                                                                     */
                                                                    faread_flags & FASTA_MAPCDSEG ? farec->cdseg[facds->seg_idx].a + s : s,
                                                                    tri_types[t], result);

                                        if (search_ret != 0) {
                                                dP("DEBUG: ret=%d\n", search_ret);
                                                ret = search_ret;
#pragma omp flush(ret)
                                        }
                                }
                        }
                        /* in case of OMP support enabled, there is an implicit barrier at this point */
                }
                fasta_rec_free(farec);
        }
        atrans_free(tr);
        fasta_close(fa);

        return (ret);
}
