/*
 *  Copyright (C) 2011 Masaryk University, Faculty of Informatics
 *
 *  Author: Daniel Kopecek <xkopecek@fi.muni.cz>
 *
 *  LICENSE TERMS
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *  3. Neither the name of the Company nor the names of its contributors
 *  may be used to endorse or promote products derived from this
 *  software without specific prior written permission.
 *
 *  This software is provided ``as is'', and any express or implied
 *  warranties, including, but not limited to, the implied warranties of
 *  merchantability and fitness for a particular purpose are disclaimed.
 *  In no event shall the company or contributors be liable for any
 *  direct, indirect, incidental, special, exemplary, or consequential
 *  damages (including, but not limited to, procurement of substitute
 *  goods or services; loss of use, data, or profits; or business
 *  interruption) however caused and on any theory of liability, whether
 *  in contract, strict liability, or tort (including negligence or
 *  otherwise) arising in any way out of the use of this software, even
 *  if advised of the possibility of such damage.
 */
#ifndef GFF3_H
#define GFF3_H

#ifdef __cplusplus
extern "C" {
#endif

#include <stdio.h>
#include <stdint.h>
#ifdef _OPENMP
#include <pthread.h>
#endif
        /**
         * tag=value pair
         */
        typedef struct {
                char *tag;
                char *value;
        } gff3_attr_t;

        /**
         * The following structure represents a GFF 3
         * record. The source field is held in the main
         * db structure to save space.
         */
        typedef struct {
                char    *seq_id;
                char    *type;

                uint64_t start;
                uint64_t end;

                double   score;

                gff3_attr_t *attr_tbl;
                uint16_t     attr_cnt;

                char     strand;
                int8_t   phase;
        } gff3_rec_t;

        struct gff3_blk {
                struct gff3_blk *next;
                size_t           blk_size;
                size_t           blk_next;
                gff3_rec_t       blk_item[];
        };

        typedef struct {
                struct gff3_blk *blk0; /* first block */
                struct gff3_blk *blkN; /* last block */
                uint32_t         rcnt; /* total number of gff3 records */

                char   *source; /**< value of the GFF source field */
                char  **scache; /**< sorted strings */
                size_t  scount; /**< number of strings in the array */

#ifdef _OPENMP
                pthread_rwlock_t rwlock;
#endif
        } gff3_t;

#define GFF3_DEFAULT_BLKSIZE 4096

        /**
         * Create a new GFF 3 db and set the source field.
         */
        gff3_t *gff3_new(const char *source);

        /**
         * Free all memory used by the given db.
         */
        int gff3_free(gff3_t *g);

        /**
         * Write the given db to a file.
         */
        int gff3_write(gff3_t *g, const char *path);

        /**
         * Prepare a GFF record for insertion into the db.
         */
        gff3_rec_t *gff3_rec_prepare(gff3_rec_t *rec, uint16_t attr_cnt);

        /**
         * Add a tag=value pair to the record.
         */
        int gff3_rec_addtag(gff3_rec_t *rec, const char *tag, char *value);

        /**
         * Insert the given record into a GFF 3 db.
         */
        int gff3_rec_add(gff3_t *g, gff3_rec_t *rec);
#ifdef __cplusplus
}
#endif

#endif /* GFF3_H */
